// DOCUMENT:  SMTPConnection-Class
// VERSION:   $Revision: 1.6 $
// DATE:      $Date: 2007-11-23 21:24:41 $
// AUTHOR:    M.Beranek
// COPYRIGHT: M.Beranek

#ifndef _SMTPCONNECTION_H
#define	_SMTPCONNECTION_H

#include <string>
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include "Thread.h"

using namespace std;

namespace SmtpThread
{

    #define STATE_NOT_CONNECTED  0
    #define STATE_CONNECTED      1
    #define STATE_HELO           2
    #define STATE_MAILFROM       3
    #define STATE_RCPTTO         4
    #define STATE_DATA           5
    #define STATE_DATA_SENDING   6


    #define SA struct sockaddr
    #define SAI struct sockaddr_in



    /**
     * SMTPConnection-Class.
     *
     * This class is inherited from the Thread-class.
     * This means, you can initialize a SMTPConnection-object,
     * setup your SMTPConnection, start threading and
     * push a message to the SMTPConnection.
     * While the SMTPConnection send the message in its thread,
     * you can continue doing other stuff.
     *
     * SMTPConnection speaks plain SMTP, no ESMTP.
     *
     */
    class SMTPConnection
    {


    public:

        /** Constructor. */
        SMTPConnection();

        /** 
         * Constructor. Initializes the smtp-connection to "host".
         * @param host IP-address of the host to connect to.
         */
        SMTPConnection( string host );
        
        /** Constructor. Initializes the smtp-connection to "host"
         * and uses <code>myname</code> as hostname.
         * @param host IP-address of the host to connect to.
         * @param myname hostname for the HELO-command.
         */
        SMTPConnection( string host, string myname );
        
        /** Destructor. */
        ~SMTPConnection();

        /**
         * Definition of the possible states of an SMTP-connection.
         */
        enum SMTPState
        {
            NOT_CONNECTED,
            CONNECTED    ,
            HELO         ,
            MAILFROM     ,
            RCPTTO       ,
            DATA         ,
            DATA_SENDING
        };

        /**
         * Gets the current state of the connection.
         * Since this is an unthreaded class, this will
         * either be "SMTPState::NOT_CONNECTED" or
         * "SMTPState::CONNECTED".
         * @return state of the current connection.
         */
        SMTPState getSMTPState();

        /**
         * Sets the IP of the SMTP-host to connect to.
         * @param host IP-address of SMTP-host.
         */
        void setSMTPHost( string host );

        /**
         * Sets the port-number of the SMTP-host to connect to.
         * @param port Port-number on SMTP-host.
         */
        void setSMTPPort( int port );

        /**
         * Sets the hostname to send in the HELO-command of the SMTP-dialog.
         * @param myname HELO-hostname.
         */
        void setMyHostname( string myname );

        /**
         * Sends a mail via SMTP.
         * @param from FROM-email
         * @param to TO-email
         * @param msg Message-body of email. Should include additional headers like "Subject" etc.
         * @return status: 0 = success, -1 = error
         */
        int sendMail( string& from, string& to, string& msg );

        /**
         * When connected, disconnects from the SMTP-Server.
         */
        void disconnect();


    private:

        // State of the current SMTP-dialog.
        SMTPState _smtp;

        // IP-Address of SMTP-Host:
        string _host;

        // HELO-Name
        string _myname;

        // SMTP-Port:
        int _port;

        // Email-From:
        string _from;

        // Email-To:
        string _to;

        // Email-Body:
        string _body;

        // Socket:
        int _tcpSocket;

        // Socket-Structure:
        SAI * _remoteEnd;

        /** Initializes the socket */
        void initSocket();

        /** Closes the socket: */
        void closeSocket();

        /**
         * Reads a line from the SMTP-Server, returning the SMTP-Code and ignoring
         * the rest of the line.
         * @return SMTP-Status-Code
         */
        int readSMTPState();

        /**
         * Sends a command to the SMTP-Server.
         * @param const string msg SMTP-Command.
         */
        int sendSMTP( const string msg );

        /**
         * Sends a command to the SMTP-Server and a CRLF, indicating "end-of-command".
         * @param const string msg SMTP-Command.
         */
        int sendSMTP_CRLF( const string msg );

        /**
         * Reads a line from the SMTP-Server.
         * @return SMTP-Status
         */
        int readSMTPLine();

        /**
         * Send the stored Email.
         * The actual work is done here.
         * @return error-code, 0 = success, -1 on error
         */
        int sendSMTPMessage();

        /**
         * Initializes the socket-structures.
         */
        void initConnection();

        /**
         * Used to store the SMTP-Status-Code, 3 characters for the code and one for
         * the terminating nullbyte.
         */
        char _state[4];

        /**
         * Buffer for reading a line of max. 999 characters from the SMTP-Server.
         */
        char _buf1[ 1000 ];


        /**
         * We are reading just one character from the SMTP-Server.
         */
        char _buf2;

    };

}
#endif	/* _SMTPCONNECTION_H */

