// DOCUMENT:  ThreadedSMTPConnection-Class
// VERSION:   $Revision: 1.7 $
// DATE:      $Date: 2007-11-23 21:24:40 $
// AUTHOR:    M.Beranek
// COPYRIGHT: M.Beranek

#ifndef _THREADEDSMTPCONNECTION_H
#define	_THREADEDSMTPCONNECTION_H

#include <string>
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include "Thread.h"
#include "SMTPConnection.h"

using namespace std;

namespace SmtpThread
{
    
    
#define STATE_NOT_CONNECTED  0
#define STATE_CONNECTED      1
#define STATE_HELO           2
#define STATE_MAILFROM       3
#define STATE_RCPTTO         4
#define STATE_DATA           5
#define STATE_DATA_SENDING   6
    
    
#define SA struct sockaddr
#define SAI struct sockaddr_in
    
    /**
     * ThreadedSMTPConnection-Class.
     *
     * This class is inherited from the Thread-class.
     * This means, you can initialize a ThreadedSMTPConnection-object,
     * setup your ThreadedSMTPConnection, start threading and
     * push a message to the ThreadedSMTPConnection.
     * While the ThreadedSMTPConnection sends the message in its thread,
     * you can continue doing other stuff.
     *
     * ThreadedSMTPConnection speaks plain SMTP, no ESMTP.
     *
     */
    class ThreadedSMTPConnection : public Thread
    {
        
        
    public:
        /**
         * Default constructor.
         */
        ThreadedSMTPConnection();
        
        /**
         * Constructor. Initializes the smtp-connection to "host".
         * @param host IP-address of the host to connect to.
         */
        ThreadedSMTPConnection( string host );
        
        /**
         * Constructor. Initializes the smtp-connection to "host"
         * and uses <code>myname</code> as hostname.
         * @param host IP-address of the host to connect to.
         * @param myname hostname for the HELO-command.
         */
        ThreadedSMTPConnection( string host, string myname );
        
        /**
         * The main thread-loop.
         * Checks whether or not a mail is provided to the object
         * and sends the mail.
         * @param arg pointer to the arguments.
         * @see Thread::Execute
         */
        void Execute( void * arg );
        
        /**
         * Returns the current status of the thread.
         * @return true / false = thread is ready / is not ready to accept mssages.
         */
        bool isReady();
        
        /**
         * Returns the current state of the SMTP-dialog.
         * @return state of the current connection.
         */
        SMTPConnection::SMTPState getSMTPState();
        
        /**
         * Sets the IP-address of the SMTP-server.
         *
         * NOTE: Must be an IP-address and not a domain-name!
         *
         * @param host IP-adress of SMTP-server.
         */
        void setSMTPHost( string host );
        
        /**
         * Sets the port-number to connect to.
         * @param port Port-number of the SMTP-server.
         */
        void setSMTPPort( int port );
        
        /**
         * Sets the hostname to send within the HELO-command.
         * @param myname Hostname of the sending host.
         */
        void setMyHostname( string myname );
        
        /**
         * Send an email to the SMTP-server.
         * @param from email-address of the sender.
         * @param to email-address of the receiver.
         * @param msg email-body, including all additional headers.
         */
        void sendMail( string& from, string& to, string& msg );
        
        
    private:
        
        /** SMTP-Connection. */
        SMTPConnection _smtpCon;
        
        /** Current state of the SMTP-dialog. */
        SMTPConnection::SMTPState _smtp;
        
        /** Status of this connection. */
        bool _isReady;
        
        /** IP-address of the SMTP-server. */
        string _host;
        
        /** Hostname used in HELO-command. */
        string _myname;
        
        /** Port-number to connect to. */
        int _port;
        
        /** Status, whether we have a mail to send or not. */
        bool _hasMail;
        
        /* Email-address of sender. */
        string _from;
        
        /* Email-address of recipient. */
        string _to;
        
        /** Mailbody, including all headers. */
        string _body;
        
        
        
    };
    
}

#endif	/* _THREADEDSMTPCONNECTION_H */

